/* Copyright (C) 2022 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Volker Friese [committer], Felix Weiglhofer */
#include "ReadoutConfig.h"

#include "CbmStsAddress.h"

#include <cassert>
#include <iomanip>

using std::pair;
using std::setw;

using namespace cbm::algo;

// ---  Constructor  ------------------------------------------------------------------
sts::ReadoutConfig::ReadoutConfig(const ReadoutSetup& config) { Init(config); }
// ------------------------------------------------------------------------------------

// ---   Equipment IDs   --------------------------------------------------------------
std::vector<u16> sts::ReadoutConfig::GetEquipmentIds()
{
  std::vector<uint16_t> result;
  for (auto& entry : fReadoutConfig)
    result.push_back(entry.first);
  return result;
}
// ------------------------------------------------------------------------------------

// ---   Number of elinks for a component / equipment   -------------------------------
size_t sts::ReadoutConfig::GetNumElinks(u16 equipmentId)
{
  size_t result = 0;
  auto it       = fReadoutConfig.find(equipmentId);
  if (it != fReadoutConfig.end()) result = fReadoutConfig[equipmentId].size();
  return result;
}
// ------------------------------------------------------------------------------------

size_t sts::ReadoutConfig::GetNumElinks()
{
  size_t result = 0;
  for (auto& entry : fReadoutConfig) {
    result += entry.second.size();
  }
  return result;
}

// ---  Initialise the mapping structure   --------------------------------------------
void sts::ReadoutConfig::Init(const ReadoutSetup& config)
{
  // The readout hierarchy is: component - CROB - FEB - ASIC (elink). Each elink
  // connects one ASIC. One FEB comprises 8 ASICs and reads out one side of a module (sensor).
  // In this setup, there is only one CROB per component. The code below is formulate such
  // as to support also multiple CROBs per component. In that case, the elinks are numbered
  // consecutively within one component.

  // Constants
  const uint16_t numModules     = config.modules.size();                      // Number of modules in the setup
  const uint16_t numComp        = config.components.size();                   // Number of components
  const uint16_t numCrobPerComp = config.components.at(0).feb2module.size();  // Number of CROBs per component
  // const uint16_t numFebsPerCrob   = config.components.at(0).feb2module.at(0).size();  // Number of FEBs per CROB
  const uint16_t numAsicsPerFeb   = config.numAsicsPerFeb;  // Number of ASICs per FEB
  const uint16_t numAsicsPerMod   = 2 * numAsicsPerFeb;     // Number of ASICs per module
  const uint16_t numElinksPerCrob = config.elinks.size();   // Number of elinks per CROB

  // Constructing the map (equipmentId, eLink) -> (module, ASIC within module)
  uint16_t numElinksPerComp = numCrobPerComp * numElinksPerCrob;
  for (uint16_t compIdx = 0; compIdx < numComp; compIdx++) {
    const auto& component = config.components.at(compIdx);
    uint16_t equipment    = component.equipmentId;
    fReadoutConfig[equipment].resize(numElinksPerComp);
    for (uint16_t crobIdx = 0; crobIdx < numCrobPerComp; crobIdx++) {
      for (uint16_t elinkIdx = 0; elinkIdx < numElinksPerCrob; elinkIdx++) {

        int32_t moduleAddress = -1;
        uint16_t asicInModule = 0;
        bool isPulser         = false;

        uint16_t elinkId  = numElinksPerCrob * crobIdx + elinkIdx;  // elink within component
        const auto& elink = config.elinks.at(elinkId);

        int16_t feb = elink.toFeb;  // FEB within CROB
        if (feb != -1) {
          int16_t moduleIdx = component.feb2module[crobIdx][feb];         // Module index
          isPulser          = component.febIsPulser.at(crobIdx).at(feb);  // Pulser flag

          if (moduleIdx != -1) {
            assert(moduleIdx < numModules);
            const auto& module = config.modules.at(moduleIdx);
            moduleAddress      = module.address;                                // Module address
            bool moduleType    = module.type;                                   // 0 or 1
            int16_t moduleSide = component.feb2moduleSide[crobIdx][feb];        // 0 or 1, -1 is inactive
            int16_t febType    = (moduleType == 0 ? moduleSide : !moduleSide);  // 0 = FEB A, 1 = FEB B
            uint32_t asicIndex = (febType == 0 ? elink.toAsicFebA : elink.toAsicFebB);
            uint32_t asicInFeb = asicIndex % numAsicsPerFeb;  // ASIC number within FEB
            // Asic number is counted downward from numAsicsPerMod - 1 for p side
            asicInModule = (moduleSide == 1 ? asicInFeb : numAsicsPerMod - 1 - asicInFeb);
          }
        }
        fReadoutConfig[equipment][elinkIdx] = {moduleAddress, asicInModule, isPulser};

      }  //# elink
    }    //# CROB
  }      //# component
}
// ------------------------------------------------------------------------------------

// ---  Mapping (equimentId, elink) -> (address, ASIC)  --------------------------------
sts::ReadoutConfig::Entry sts::ReadoutConfig::Map(uint16_t equipmentId, uint16_t elinkId)
{
  Entry result {-1, 0, false};
  auto equipIter = fReadoutConfig.find(equipmentId);
  if (equipIter != fReadoutConfig.end()) {
    if (elinkId < equipIter->second.size()) { result = equipIter->second.at(elinkId); }
  }
  return result;
}
// ------------------------------------------------------------------------------------

// -----   Print readout map   ------------------------------------------------
std::string sts::ReadoutConfig::PrintReadoutMap()
{

  std::stringstream ss;
  for (auto& equipment : fReadoutConfig) {
    auto eqId = equipment.first;
    for (size_t elink = 0; elink < equipment.second.size(); elink++) {
      auto address = equipment.second.at(elink).moduleAddress;
      auto asicNr  = equipment.second.at(elink).asicNumber;
      ss << "\n Equipment " << eqId << "  elink " << setw(2) << elink;
      ss << "  ASIC " << setw(2) << asicNr << "  module " << address;
      if (address != -1) {
        ss << "  Unit " << setw(2) << CbmStsAddress::GetElementId(address, kStsUnit);
        ss << "  Ladd " << setw(2) << CbmStsAddress::GetElementId(address, kStsLadder);
        ss << "  Hlad " << setw(2) << CbmStsAddress::GetElementId(address, kStsHalfLadder);
        ss << "  Modu " << setw(2) << CbmStsAddress::GetElementId(address, kStsModule);
      }
      else
        ss << "  Inactive";
    }  //# elink
  }    //# component
  return ss.str();
}
// ----------------------------------------------------------------------------
